/*
cTrans.h

Transformation Functions:
  sk: gives a skew symmetric matrix from a given vector w
  llarate: rate of change of latitude, longitude, and altitude
  lla2ecef: LLA to ECEF
  ecef2lla: ECEF to LLA
  body2ned: IMU body to NED inertial
  ecef2ned: ECEF to NED
  quat2dcm: quaternions to DCM
  ned2ecef: NED to ECEF
  qmult: quaternion multiply
  constrainAngle180: constrains angle to +- 180 degrees
  constrainAngle360: constrains angle from 0 to 360 degrees
  
Original Author: 

Brian R Taylor
brian.taylor@bolderflight.com
2017-12-20
Bolder Flight Systems
Copyright 2017 Bolder Flight Systems

Permission is hereby granted, free of charge, to any person obtaining a copy of this software 
and associated documentation files (the "Software"), to deal in the Software without restriction, 
including without limitation the rights to use, copy, modify, merge, publish, distribute, 
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or 
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING 
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _cTrans_
#define _cTrans_

#include "Arduino.h"
#include "Eigen.h"
#include <Eigen/Dense>

// acceleration due to gravity
const double gravityConst = 9.80665f;
// major eccentricity squared
const double ECC2 = 0.0066943799901;
// earth semi-major axis radius (m)
const double EARTH_RADIUS = 6378137.0;  

// This function gives a skew symmetric matrix from a given vector w
Eigen::Matrix<double,3,3> sk(Eigen::Matrix<double,3,1> w) {
  Eigen::Matrix<double,3,3> C;
  C(0,0) = 0.0f;    C(0,1) = -w(2,0); C(0,2) = w(1,0);
  C(1,0) = w(2,0);  C(1,1) = 0.0f;    C(1,2) = -w(0,0);
  C(2,0) = -w(1,0); C(2,1) = w(0,0);  C(2,2) = 0.0f;
  return C;
}

// This function calculates the rate of change of latitude, longitude, and altitude.
Eigen::Matrix<double,3,1> llarate(Eigen::Matrix<double,3,1> V,Eigen::Matrix<double,3,1> lla) {
  double Rew, Rns, denom;
  Eigen::Matrix<double,3,1> lla_dot;
  
  denom = (1.0 - (ECC2 * pow(sin(lla(0,0)),2.0)));
  denom = sqrt(denom*denom);

  Rew = EARTH_RADIUS / sqrt(denom);
  Rns = EARTH_RADIUS*(1.0-ECC2) / denom*sqrt(denom);
  
  lla_dot(0,0) = V(0,0)/(Rns + lla(2,0));
  lla_dot(1,0) = V(1,0)/((Rew + lla(2,0))*cos(lla(0,0)));
  lla_dot(2,0) = -V(2,0);
  
  return lla_dot;
}

// This function calculates the ECEF Coordinate given the Latitude, Longitude and Altitude.
Eigen::Matrix<double,3,1> lla2ecef(Eigen::Matrix<double,3,1> lla) {
  double Rew, denom;
  Eigen::Matrix<double,3,1> ecef;

  denom = (1.0 - (ECC2 * pow(sin(lla(0,0)),2.0)));
  denom = sqrt(denom*denom);

  Rew = EARTH_RADIUS / sqrt(denom);

  ecef(0,0) = (Rew + lla(2,0)) * cos(lla(0,0)) * cos(lla(1,0));
  ecef(1,0) = (Rew + lla(2,0)) * cos(lla(0,0)) * sin(lla(1,0));
  ecef(2,0) = (Rew * (1.0 - ECC2) + lla(2,0)) * sin(lla(0,0));

  return ecef;
}

// This function calculates the Latitude, Longitude and Altitude given ECEF Coordinates.
Eigen::Matrix<double,3,1> ecef2lla(Eigen::Matrix<double,3,1> ecef) {
	
  double a,f,b,e,ePrime,n,lat,lon,alt,p,theta,x,y,z;
  Eigen::Matrix<double,3,1> lla;

  // WGS84 parameters
  //a = 6378137.0;
  a = EARTH_RADIUS;
  f = 1./298.257223563;
  b = a*(1.-f);
  e = sqrt( (a*a - b*b) / (a*a) );
  ePrime = sqrt( (a*a - b*b) / (b*b) );
  
  // local variables
  x = ecef(0,0);
  y = ecef(1,0);
  z = ecef(2,0);

  // calculations
  lon = atan2(y,x);
  p = sqrt(x*x + y*y);
  theta = atan2(a*z,b*p);
  lat = atan2((z+ePrime*ePrime*b*pow(sin(theta),3.0)),\
              (p-e*e*a*pow(cos(theta),3.0)));
  n = a / (sqrt(1. - e*e*pow(sin(lat),2.0)));           
  alt = (p/(cos(lat)))-n;

  // correct for numerical instability near poles
  //k = abs(x)<1 & abs(y)<1;
  //alt(k) = abs(z(k))-b;
  
  if(lon>PI){
	  lon = lon - 2*PI;
  }
 
  lla(0,0) = lat;
  lla(1,0) = lon;
  lla(2,0) = alt;

  return lla;
}

// This function calculates the NED vector given body vector.
//   ypr is a vector of yaw, pitch, and roll in radians
Eigen::Matrix<double,3,1> body2ned(Eigen::Matrix<double,3,1> ypr,Eigen::Matrix<double,3,1> ecef) {
	
  double pR, rR, yR;
  Eigen::Matrix<double,3,1> ned;
  Eigen::Matrix<double,3,3> dcm;

  // temp variables
  pR = ypr(1,0);
  rR = ypr(2,0);
  yR = ypr(0,0);

  // define DCM 
  dcm(0,0) =  cos(pR) * cos(yR);
  dcm(0,1) =  cos(pR) * sin(yR);
  dcm(0,2) = -sin(pR);

  dcm(1,0) = -cos(rR) * sin(yR) + sin(rR) * sin(pR) * cos(yR) ;
  dcm(1,1) =  cos(rR) * cos(yR) + sin(rR) * sin(pR) * sin(yR);
  dcm(1,2) =  sin(rR) * cos(pR);

  dcm(2,0) =  sin(rR) * sin(yR) + cos(rR) * sin(pR) * cos(yR);
  dcm(2,1) = -sin(rR) * cos(yR) + cos(rR) * sin(pR) * sin(yR) ;
  dcm(2,2) =  cos(rR) * cos(pR);

  // Perform transformation
  ned = dcm.transpose() * ecef;

  // Remove gravity from Z component
  ned(2,0) = ned(2,0) + gravityConst;
  
  return ned;
}

// This function converts a vector in ecef to ned coordinate centered at pos_ref.
Eigen::Matrix<double,3,1> ecef2ned(Eigen::Matrix<double,3,1> ecef,Eigen::Matrix<double,3,1> pos_ref) {
  Eigen::Matrix<double,3,1> ned;
  ned(2,0)=-cos(pos_ref(0,0))*cos(pos_ref(1,0))*ecef(0,0)-cos(pos_ref(0,0))*sin(pos_ref(1,0))*ecef(1,0)-sin(pos_ref(0,0))*ecef(2,0);
  ned(1,0)=-sin(pos_ref(1,0))*ecef(0,0) + cos(pos_ref(1,0))*ecef(1,0);
  ned(0,0)=-sin(pos_ref(0,0))*cos(pos_ref(1,0))*ecef(0,0)-sin(pos_ref(0,0))*sin(pos_ref(1,0))*ecef(1,0)+cos(pos_ref(0,0))*ecef(2,0);
  return ned;
}

// quaternion to dcm
Eigen::Matrix<double,3,3> quat2dcm(Eigen::Matrix<double,4,1> q) {
  Eigen::Matrix<double,3,3> C_N2B;
  C_N2B(0,0) = 2.0f*powf(q(0,0),2.0f)-1.0f + 2.0f*powf(q(1,0),2.0f);
  C_N2B(1,1) = 2.0f*powf(q(0,0),2.0f)-1.0f + 2.0f*powf(q(2,0),2.0f);
  C_N2B(2,2) = 2.0f*powf(q(0,0),2.0f)-1.0f + 2.0f*powf(q(3,0),2.0f);
  
  C_N2B(0,1) = 2.0f*q(1,0)*q(2,0) + 2.0f*q(0,0)*q(3,0);
  C_N2B(0,2) = 2.0f*q(1,0)*q(3,0) - 2.0f*q(0,0)*q(2,0);
  
  C_N2B(1,0) = 2.0f*q(1,0)*q(2,0) - 2.0f*q(0,0)*q(3,0);
  C_N2B(1,2) = 2.0f*q(2,0)*q(3,0) + 2.0f*q(0,0)*q(1,0);
  
  C_N2B(2,0) = 2.0f*q(1,0)*q(3,0) + 2.0f*q(0,0)*q(2,0);
  C_N2B(2,1) = 2.0f*q(2,0)*q(3,0) - 2.0f*q(0,0)*q(1,0);
  return C_N2B;
}

// ned to ecef inertial
//   ned is the input NED vector
//   refLLA is vector of lat (rad), lon (rad), alt (m above ellipsoid)
Eigen::Matrix<double,3,1> ned2ecef(Eigen::Matrix<double,3,1> refLLA,Eigen::Matrix<double,3,1> ned) {
	
  Eigen::Matrix<double,3,3> C_N2I;
  Eigen::Matrix<double,3,1> ecef;
  double clat, slat, clon, slon;
  
  // Temporary variables
  clat = cos(refLLA(0,0));
  slat = sin(refLLA(0,0));
  clon = cos(refLLA(1,0));
  slon = sin(refLLA(1,0));
  
  C_N2I(0,0) = -clon * slat;
  C_N2I(0,1) = -slon;
  C_N2I(0,2) =  clon * clat;
  
  C_N2I(1,0) = -slon * slat;
  C_N2I(1,1) =  clon;
  C_N2I(1,2) = -slon * clat;
  
  C_N2I(2,0) =  clat;
  C_N2I(2,1) =  0.0f;
  C_N2I(2,2) = -slat;

  ecef = C_N2I * ned;
  return ecef;
}

// quaternion multiplication
Eigen::Matrix<double,4,1> qmult(Eigen::Matrix<double,4,1> p, Eigen::Matrix<double,4,1> q) {
  Eigen::Matrix<double,4,1> r;
  r(0,0) = p(0,0)*q(0,0) - (p(1,0)*q(1,0) + p(2,0)*q(2,0) + p(3,0)*q(3,0));
  r(1,0) = p(0,0)*q(1,0) + q(0,0)*p(1,0) + p(2,0)*q(3,0) - p(3,0)*q(2,0);
  r(2,0) = p(0,0)*q(2,0) + q(0,0)*p(2,0) + p(3,0)*q(1,0) - p(1,0)*q(3,0);
  r(3,0) = p(0,0)*q(3,0) + q(0,0)*p(3,0) + p(1,0)*q(2,0) - p(2,0)*q(1,0);
  return r;
}

// bound yaw angle between -180 and 180
double constrainAngle180(double dta) {
  if(dta >  PI) dta -= (PI*2.0f);
  if(dta < -PI) dta += (PI*2.0f);
  return dta;
}

// bound heading angle between 0 and 360 
double constrainAngle360(double dta){
  dta = fmod(dta,2.0f*PI);
  if (dta < 0)
    dta += 2.0f*PI;
  return dta;
}

#endif


