///////////////////////////////////////////////////////////////////////////////////////
// 
// Kalman filter INS routines
///////////////////////////////////////////////////////////////////////////////////////

// Kalman filter initialization
void kfInit(){

  // Initialize R noise covariance
  rMatrix(0,0) = pow(rPosnSigma, 2.0);
  rMatrix(1,1) = pow(rPosnSigma, 2.0);
  rMatrix(2,2) = pow(rPosnSigma, 2.0);
  rMatrix(3,3) = pow(rVelSigma, 2.0);
  rMatrix(4,4) = pow(rVelSigma, 2.0);
  rMatrix(5,5) = pow(rVelSigma, 2.0);

  // Initialize state and covariance, x and P
  // Use first LLA GPS reading to initialize x

  pk(0,0) = pow(initPosnVar, 2.0);
  pk(1,1) = pow(initPosnVar, 2.0);
  pk(2,2) = pow(initPosnVar, 2.0);
  pk(3,3) = pow(initVelVar, 2.0);
  pk(4,4) = pow(initVelVar, 2.0);
  pk(5,5) = pow(initVelVar, 2.0);
  pk(6,6) = pow(initAccVar, 2.0);
  pk(7,7) = pow(initAccVar, 2.0);
  pk(8,8) = pow(initAccVar, 2.0);

  RefLLA(0,0) = double(rtk.lat1 * deg2rad);
  RefLLA(1,0) = double(rtk.lon1 * deg2rad);
  RefLLA(2,0) = double(rtk.hMSL);  
  xECEF = lla2ecef(RefLLA);
  xk(0,0) = xECEF(0,0);
  xk(1,0) = xECEF(1,0);
  xk(2,0) = xECEF(2,0);

  // Initialize appropriate H, F, and G elements to identity
  hMatrix(0,0) = 1.0;
  hMatrix(1,1) = 1.0;
  hMatrix(2,2) = 1.0;
  hMatrix(3,3) = 1.0;
  hMatrix(4,4) = 1.0;
  hMatrix(5,5) = 1.0;
  fMatrix = Eigen::Matrix<double,9,9>::Identity();
  gMatrix(6,0) = 1.0;
  gMatrix(7,1) = 1.0;
  gMatrix(8,2) = 1.0;

  // Print out initial GPS LLA, filter LLA, filter ECEF
  Serial.print("uBlox LLA: ");
  Serial.print(rtk.lat1);
  Serial.print(",  ");
  Serial.print(rtk.lon1);
  Serial.print(",  ");
  Serial.println(rtk.hMSL);

  Serial.print("RefLLA Initial: ");
  Serial.print(RefLLA(0,0)*rad2deg);
  Serial.print(",  ");
  Serial.print(RefLLA(1,0)*rad2deg);
  Serial.print(",  ");
  Serial.println(RefLLA(2,0));
    
  Serial.print("xk Initial: ");
  Serial.print(xk(0,0));
  Serial.print(",  ");
  Serial.print(xk(1,0));
  Serial.print(",  ");
  Serial.println(xk(2,0));
  Serial.println(" ");
}

// Kalman filter propagation
void kfProp(){

  // Convert IMU accel into ECEF. Note that we'll 
  // use the previous accel (time k) to propagate
  // the state to time k1.
  accNEDk1(0,0) = aNEDx1;
  accNEDk1(1,0) = aNEDy1;
  accNEDk1(2,0) = aNEDz1;
  RefLLA(0,0) = double(rtk.lat1 * deg2rad);
  RefLLA(1,0) = double(rtk.lon1 * deg2rad);
  RefLLA(2,0) = double(rtk.hMSL);  
  accECEFk1 = ned2ecef(RefLLA,accNEDk1);

  // Use timestep to update F, G, and Q
  fMatrix(0,3) = _dt;
  fMatrix(1,4) = _dt;
  fMatrix(2,5) = _dt;
  fMatrix(3,6) = _dt;
  fMatrix(4,7) = _dt;
  fMatrix(5,8) = _dt;
  fMatrix(0,6) = 0.5 * _dt * _dt;
  fMatrix(1,7) = 0.5 * _dt * _dt;
  fMatrix(2,8) = 0.5 * _dt * _dt;
  
  gMatrix(0,0) = 0.5 * _dt * _dt;
  gMatrix(1,1) = 0.5 * _dt * _dt;
  gMatrix(2,2) = 0.5 * _dt * _dt;
  gMatrix(3,0) = _dt;
  gMatrix(4,1) = _dt;
  gMatrix(5,2) = _dt;

  calculateQ();

  // Perform Kalman propagation.
  //   Use first line to include IMU accel as control inputs
  //   Use second line to not include IMU accel as control inputs
  xk1Minus = fMatrix * xk + gMatrix * accECEFk; 
  //xk1Minus = fMatrix * xk;
  pk1Minus = fMatrix * pk * fMatrix.transpose() + qMatrix;
}

// Kalman filter measurement update
void kfUpdate(){
  // Calculate Kalman gain
  kk1 = pk1Minus * hMatrix.transpose() * (hMatrix * pk1Minus * hMatrix.transpose() + rMatrix).inverse();
        
  // read in GPS position and velocity measurements
  RefLLA(0,0) = double(rtk.lat1 * deg2rad);
  RefLLA(1,0) = double(rtk.lon1 * deg2rad);
  RefLLA(2,0) = double(rtk.hMSL);  
  velNEDk1(0,0) = rtk.velN;
  velNEDk1(1,0) = rtk.velE;
  velNEDk1(2,0) = rtk.velD;

  // Convert measurements to ECEF
  posECEFk1 = lla2ecef(RefLLA);
  velECEFk1 = ned2ecef(RefLLA,velNEDk1);

  // Load measurements into measurement vector zk1
  zk1(0,0) = posECEFk1(0,0);
  zk1(1,0) = posECEFk1(1,0);
  zk1(2,0) = posECEFk1(2,0);
  zk1(3,0) = velECEFk1(0,0);
  zk1(4,0) = velECEFk1(1,0);
  zk1(5,0) = velECEFk1(2,0);

  // Calculate innovations
  nuk1 = zk1 - hMatrix * xk1Minus;

  // Perform Kalman update
  xk1 = xk1Minus + kk1 * nuk1;
  pk1 = (Eigen::Matrix<double,9,9>::Identity() - kk1 * hMatrix) * pk1Minus;

  // Set updateFlag to true so kfReset knows a KF update occurred.
  updateFlag = true;
}

// Reset certain KF terms for next iteration
void kfReset(){
  // if KF update was performed, reset xk and pk to k1 terms
  if(updateFlag){
    xk = xk1;
    pk = pk1;
    // reset updateFlag to false
    updateFlag = false;
  // else reset xk and pk to k1Minus terms
  }else{
    xk1 = xk1Minus;
    xk = xk1Minus;
    pk1 = pk1Minus;
    pk = pk1;
  }

  // reset accel
  accECEFk = accECEFk1;
}

//#endif


